from .exceptions import AssignError

class StrictClass:
	""" StrictClass is used to compare the keyword arguments datatypes with the desired datatypes. """

	def __init__(self, **kwargs) -> None:
		""" Initialize object attributes.

		:param kwargs: everything that must be validated and/or stored inside the object """

		# Check if there is any annotation.
		if '__annotations__' in dir(self):

			# Check each annotation key and the desired datatype with the input parameters.
			for key in self.__annotations__:

				# Check if every annotation key was passed in the parameters.
				if key in kwargs:

					# Has the annotation more than one expression? ...
					if '__args__' in dir(self.__annotations__[key]):

						# Does the keyword object's datatype match any of the annotation expression datatypes? So raise exception if not.
						if not type(kwargs[key]) in self.__annotations__[key].__args__:
							raise AssignError(
								'Trying to assign an invalid datatype to the parameter "%s". Datatype must be %s, and not %s.'
								% (key, " or ".join([str(t) for t in self.__annotations__[key].__args__]), type(kwargs[key])),
								parameter=key, required=self.__annotations__[key].__args__, invalid=type(kwargs[key])
							)

					# Does it only has one expression? So raise exception if the datatypes don't match.
					elif not type(kwargs[key]) is self.__annotations__[key]:
						raise AssignError(
							'Trying to assign an invalid datatype to the parameter "%s". Datatype must be %s, and not %s.'
							% (key, self.__annotations__[key], type(kwargs[key])),
							parameter=key, required=(self.__annotations__[key],), invalid=type(kwargs[key])
						)

				# Raise exception if any annotations parameter isn't available.
				else:
					raise AssignError('A missing keyword ("%s") must be passed.' % key, parameter=key)

		# After the annotations validations, store every keyword inside the object.
		for keyword in kwargs:
			self.__dict__[keyword] = kwargs[keyword]

	def __getitem__(self, name):
		return self.__dict__[name]

	def __iter__(self):
		return iter(self.__dict__)


class AssignError(Exception):
	""" AssignError will be raised when an invalid datatype is assigned to an attribute. """

	def __init__(self, message, parameter=None, required=None, invalid=None):
		self.parameter = parameter
		self.required = required
		self.invalid = invalid

		super().__init__(message)


""" Python strict datatypes for functions and objects. Check if the datatypes of the input parameters match the datatypes you want.
(GitHub) Written by: badlock """

from .strict_class import StrictClass
from .strict_function import StrictFunction
from .strict_type import strict_type
from .exceptions import AssignError



from .strict_class import StrictClass

class StrictFunction(StrictClass):
	""" StrictFunction is used to compare the keyword arguments datatypes with the desired datatypes and run a function
	with those arguments. """

	def __init__(self, func, **kwargs):
		""" Initialize object attributes.

		:param func: the function this object should execute
		:param kwargs: everything that must be validated and stored inside the object before calling the function """

		# Validate every keyword.
		super().__init__(**kwargs)

		# Run the function with the keywords and stores the output of the function.
		self.__function = lambda: func(**kwargs)

	def __call__(self):
		return self.__function()


from setuptools import setup

with open('README.md', 'r') as f:
    setup(
        name='strict-types',
        version='1.2',
        description='Type checker for functions and classes.',
        long_description=f.read(),
        long_description_content_type='text/markdown',
        license='MIT',
        url='https://github.com/badlock/strict-types',
        author='Badlock',
        packages=['strict'],
        package_dir={'': 'src'}
    )


from .exceptions import AssignError

def strict_type(func):
	""" strict_type is a decorator that will check the desired datatypes of a function and compare with its
	parameters datatypes. """

	def wrap(*args, **kwargs):

		varnames = func.__code__.co_varnames[:func.__code__.co_argcount]

		if func.__defaults__:
			default_arguments = dict(zip(varnames[len(func.__defaults__) * -1:], func.__defaults__))

			for darg in default_arguments:
				if darg not in kwargs:
					kwargs[darg] = default_arguments[darg]

			del default_arguments

		# Go through every argument of the function (in order).
		for n, arg in enumerate(varnames):

			# Store the parameter datatype, being it an argument or a keyword argument.
			compare_type = type(kwargs[arg]) if arg in kwargs else type(args[n])

			# Ignore if the argument doesn't have any annotation.
			# Otherwise, compare the annotation datatype with that of the parameter. If it doesn't match, raise an exception.
			if arg in func.__annotations__ and not compare_type in (func.__annotations__[arg].__args__ if '__args__' in dir(func.__annotations__[arg]) else (func.__annotations__[arg],)):
				raise AssignError(
					'Trying to assign an invalid datatype to the parameter "%s". Datatype must be %s, and not %s.'
					% (arg, func.__annotations__[arg] if not '__args__' in dir(func.__annotations__[arg]) else " or ".join([str(t) for t in func.__annotations__[arg].__args__]), compare_type),
					parameter=arg, required=func.__annotations__[arg].__args__ if '__args__' in dir(func.__annotations__[arg]) else (func.__annotations__[arg],), invalid=compare_type,
				)

		return func(*args, **kwargs)
	return wrap
